/* $Id: nfs_proc.c,v 1.24 1997/07/12 00:00:13 gjb Exp $
 * nfs_proc.c - /proc/net/nfs-{cache,mounts,profile,usage} pseudo-files
 * (C) 1997 Greg J. Badros -- All Rights Reserved
 * <gjb@cs.washington.edu> 
 * Seattle, WA  USA
 * http://www.cs.washington.edu/homes/gjb
 */

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/locks.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
#include <linux/sunrpc/clnt.h>
#include <linux/nfs_fs.h>
#include <linux/lockd/bind.h>

#include <asm/system.h>
#ifdef KERNEL_2_1
#include <asm/uaccess.h>
#else
#define copy_to_user(to, from, size) memcpy_tofs(to, from, size)
#define lockd_up() ;
#define lockd_down() ;
#endif

#include <linux/proc_fs.h>
#include <linux/file.h>
#include "dir.h"
#include "profile.h"
#include "nfs_cache.h"
#include "cache_failed_lookups.h"

#define NFS_MAX_BITMAP_SIZE_TO_SHOW 1500 /* pages, 4k pages means this is 6.1MB */

#define PROFILE_LINE_FOR(x) len += sprintf(buf+len,"%d: " #x "\t%6ld\n",i_ ## x, c_nfs_ops[i_ ## x])
#define PROFILE_LINE_FOR_RPC(x) irpc = i_rpc_ ## x; len += sprintf(buf+len,"%d: " #x "\t%6ld\n", irpc, c_nfs_ops[irpc])
#define PROFILE_LINE_FOR_MISC(x) irpc = i_ ## x; len += sprintf(buf+len,"%d: " #x "\t%6ld\n", irpc, c_nfs_ops[irpc])

extern char szNFS_Identifier[];

#ifdef MODULE
extern long mod_use_count_;
#endif

extern unsigned long cRPCMinorTimeouts;
extern unsigned long cRPCMajorTimeouts;

/* using drivers/char/misc.c's proc_misc_read as an example */
int nfs_proc_get_info(char *buf, char **start, off_t offset, int len, int unused)
{
	int irpc = 0;

	len = 0;
	len += sprintf(buf+len,"NFS module. $Date: 1997/07/12 00:00:13 $ $Revision: 1.24 $\n");
	len += sprintf(buf+len,"%s\n",szNFS_Identifier);
	len += sprintf(buf+len,"RPC: Minor, major timeouts = %lu, %lu\n",
								 cRPCMinorTimeouts, cRPCMajorTimeouts);
	len += sprintf(buf+len,"\nSuper block operations:\n");
	PROFILE_LINE_FOR(read_inode);
	PROFILE_LINE_FOR(notify_change);
	PROFILE_LINE_FOR(put_inode);
	PROFILE_LINE_FOR(put_super);
	PROFILE_LINE_FOR(statfs);

	len += sprintf(buf+len,"\nDir operations:\n");
	PROFILE_LINE_FOR(readdir);
	PROFILE_LINE_FOR(dir_open);

	len += sprintf(buf+len,"\nInode (Dir/File) operations:\n");
	PROFILE_LINE_FOR(create);
	PROFILE_LINE_FOR(lookup);
	PROFILE_LINE_FOR(link);
	PROFILE_LINE_FOR(unlink);
	PROFILE_LINE_FOR(symlink);
	PROFILE_LINE_FOR(mkdir);
	PROFILE_LINE_FOR(rmdir);
	PROFILE_LINE_FOR(mknod);
	PROFILE_LINE_FOR(rename);

	len += sprintf(buf+len,"\nFile operations:\n");
	PROFILE_LINE_FOR(file_read);
	PROFILE_LINE_FOR(file_write);
	PROFILE_LINE_FOR(file_mmap);
	PROFILE_LINE_FOR(fsync);
	PROFILE_LINE_FOR(readpage);
	
	len += sprintf(buf+len,"\nRPCs (call, alloc_fail, ok, retry, fail)\n");
	
	PROFILE_LINE_FOR_RPC(do_nothing);
	PROFILE_LINE_FOR_RPC(getattr);
	PROFILE_LINE_FOR_RPC(setattr);
	PROFILE_LINE_FOR_RPC(lookup);
	PROFILE_LINE_FOR_MISC(lookup_not_found);
	PROFILE_LINE_FOR_RPC(readlink);
	PROFILE_LINE_FOR_RPC(read);
	len += sprintf(buf+len,"Synch, Asynch reads:\t%6ld  %6ld\n",
		       c_nfs_ops[i_readpage_sync],
		       c_nfs_ops[i_readpage_async]);
	len += sprintf(buf+len,"Read fails, no_enque:\t%6ld  %6ld\n",
		       c_nfs_ops[i_readpage_async_fail],c_nfs_ops[i_readpage_async_no_enque]);
	len += sprintf(buf+len,"readpage:local,total; hit%%:\t%6ld /%6ld = %d%%\n\n",
								 c_nfs_ops[i_readpage_local],c_nfs_ops[i_readpage_total],
								 c_nfs_ops[i_readpage_local]*100/
								 (c_nfs_ops[i_readpage_total]?c_nfs_ops[i_readpage_total]:1));
	PROFILE_LINE_FOR_RPC(write);
  len += sprintf(buf+len,"Asynch writes, fails:\t%6ld  %6ld\n",
								 c_nfs_ops[i_create_write_request],c_nfs_ops[i_write_async_fail]);
	PROFILE_LINE_FOR_RPC(create);
	PROFILE_LINE_FOR_RPC(remove);
	PROFILE_LINE_FOR_RPC(rename);
	PROFILE_LINE_FOR_RPC(link);
	PROFILE_LINE_FOR_RPC(symlink);
	PROFILE_LINE_FOR_RPC(mkdir);
	PROFILE_LINE_FOR_RPC(rmdir);
	PROFILE_LINE_FOR_RPC(readdir);
	PROFILE_LINE_FOR_RPC(statfs);

	len += sprintf(buf+len,"\nLookup cache operations\n");
	PROFILE_LINE_FOR(lookup);
	PROFILE_LINE_FOR(lookup_miss);
	PROFILE_LINE_FOR_MISC(lookup_cache_index);
	PROFILE_LINE_FOR_MISC(lookup_cache_remove);
	PROFILE_LINE_FOR_MISC(lookup_cache_refresh);

	len += sprintf(buf+len,"\nLookup predictions\n");
	len += sprintf(buf+len,"Made,Correct,Unused\t%6ld  %6ld  %6ld\n",
								 cPredMadeTotal, cPredCorrect, cPredUnused);
	len += sprintf(buf+len,"PredHits,RecentUseHits,TotalLookups\t%6ld  %6ld  %6ld\n",
								 cPredHits,cRecentUseHit,cTotLookups);

	return len;
}

#undef PROFILE_LINE_FOR
#undef PROFILE_LINE_FOR_RPC
#undef PROFILE_LINE_FOR_MISC

int
string_of_bitmap_for(char *sz,const struct inode_being_cached_info *const pbci,
										 int maxlen)
{
	char *pchBitmap = pbci->pchValidBitmap;
	int cpages = pbci->cpagesTotal;
	int i = 0;

	if (cpages > NFS_MAX_BITMAP_SIZE_TO_SHOW)
		return sprintf(sz,"Bitmap too large\n");

	/* FIXGJB: may want to output in packed encoding -- make that
		 a compile time option, perhaps */
	for (;i<cpages && i<maxlen;i++) {
		sz[i] = nfs_is_bitmap_valid_for_ipage(pchBitmap,i)? '*':'-';
	}
	if (i>=maxlen) {
		i+=sprintf(sz+i,"...");
	}
	sz[i++] = '\n';
	return i;
}

int nfs_cache_proc_get_info(char *buf, char **start, off_t offset, int len, int unused)
{
	struct inode *inode_first = get_first_inode();
	struct inode *inodeCache;
	struct inode_being_cached_info *pbci;
	int i=0;
	int valid;
	int total;
	int fIsInMemory;
	len = 0;
	len += sprintf(buf+len,"NFS module. $Date: 1997/07/12 00:00:13 $ $Revision: 1.24 $\n");
	len += sprintf(buf+len,"%s\n",szNFS_Identifier);

	verb_nfs_printk("NFS: nfs_cache_proc_get_info, first_inode = %lx\n",
									(long unsigned) inode_first);
	for(inodeCache = inode_first; inodeCache; i++, inodeCache = inodeCache->i_next) {
		/* be sure we do not loop forever! FIXGJB, we'd like the
		 actually # of inodes nr_inodes, but that's not exported now */
		if (i>NR_INODE*8) {
			printk("NFS: nfs_cache_proc_get_info, endless loop\n");
			break;
		}

		if (inodeCache == NULL)
			break;
		if (inodeCache == inode_first && i>0)
			break;;

		if (inodeCache->i_count == 0)
			continue;
		if ((pbci = inodeCache->i_ibcinfo) != NULL) {
			char *szHostname = "NOSERVER";
			if (pbci->nfs_sb_info)
				szHostname = pbci->nfs_sb_info->s_server.hostname;
			valid = pbci->cpagesValid;
			total = pbci->cpagesTotal;
			fIsInMemory = pbci->fInodeNFSInMemory;
			/* use parens for in-memory, [] for out of memory */
			len += sprintf(buf+len,"%s %c%d,%ld%c -> (%d,%ld), %d, %d/%d%s\n",
										 szHostname,
										 fIsInMemory?'(':'[',
										 pbci->device, pbci->inode_number, 
										 fIsInMemory?')':']',
										 inodeCache->i_dev,inodeCache->i_ino,
										 inodeCache->i_count,
										 valid,total,
										 pbci->pchValidBitmap?"":" complete");
			if (pbci->pchValidBitmap)
				len += string_of_bitmap_for(buf+len,pbci,PAGE_SIZE-len-95);
			verb_nfs_printk("NFS: nfs_cache_proc_get_info, len = %d\n", len);
			/* be sure we do not overflow the buffer; buffer size is one page */
			if (len >= PAGE_SIZE-100) {
				printk("NFS: nfs_cache_proc_get_info: possibly incomplete listing\n");
				len += sprintf(buf+len,"...incomplete listing...\n");
				break;
			}
		}

	}

	return len;
}

int nfs_usage_proc_get_info(char *buf, char **start, off_t offset, int len, int unused)
{
	struct inode *inode_first = get_first_inode();
	struct inode *inode;
	int i=0;
	len = 0;
	len += sprintf(buf+len,"NFS module. $Date: 1997/07/12 00:00:13 $ $Revision: 1.24 $\n");
	len += sprintf(buf+len,"%s\n",szNFS_Identifier);

	for(inode = inode_first; inode; i++, inode = inode->i_next) {
		if (inode == NULL)
			break;
		if (inode->i_sb->s_magic == NFS_SUPER_MAGIC && inode->i_count > 0) {
			struct inode *inodeCache;
			inodeCache = NFS_CACHE_INODE(inode);
			len += sprintf(buf+len,"(%d,%ld) -> (%d,%d) has count = %d\n",
										 inode->i_dev, inode->i_ino,
										 inodeCache? inodeCache->i_dev:-1,
										 inodeCache? inodeCache->i_ino:-1,
										 inode->i_count);
			if (len >= PAGE_SIZE-100) {
				len += sprintf(buf+len,"...incomplete listing...\n");
				return len;
			}
		}
		/* be sure we do not loop forever! FIXGJB, we'd like the
		 actually # of inodes nr_inodes, but that's not exported now */
		if (i>NR_INODE*8) 
			return len;
		if (inode == inode_first && i>0)
			return len;
	}

	return len;
}

extern int nfs_fillind_paused;

int nfs_mounts_proc_get_info(char *buf, char **start, off_t offset, int len, int unused)
{
	struct super_block *sb;
	struct nfs_sb_info *nfs_sbi; 
	struct nfs_cache *nfs_cache;
	struct nfs_server *nfss;
	len = 0;
	len += sprintf(buf+len,"NFS module. $Date: 1997/07/12 00:00:13 $ $Revision: 1.24 $\n");
	len += sprintf(buf+len,"%s\n",szNFS_Identifier);
	len += sprintf(buf+len,"nfsfillind: %s\n",nfs_fillind_paused?"paused":"running");
	
	for (sb = super_blocks + 0 ; sb < super_blocks + NR_SUPER ; sb++) {
		if (!sb->s_dev)
			continue;
		if (sb->s_magic != NFS_SUPER_MAGIC)
			continue;
		nfs_sbi = &sb->u.nfs_sb;
		nfs_cache = nfs_sbi->s_cache;
		nfss = &nfs_sbi->s_server;
		/* FIXGJB: want mnt_devname, mnt_dirname from the vfsmount struct,
			 perhaps we need to have a back pointer from superblock to the two names,
			 alternatively, make the vfsmount linked list exported, and loop through
			 it, more like is done for PROC_MTAB in get_filesystem_info in proc/super.c */
		len += sprintf(buf+len,"%s,%d: ", 
									 nfss->hostname,sb->s_dev);
		if (nfs_cache) {
			len += sprintf(buf+len,"has %d/%d pages used on %s",
										 nfs_sbi->dc_pages_used,nfs_sbi->dc_max_size,
										 nfs_cache->root);
			len += sprintf(buf+len,"\n#  min_free=%d,max_space=%d",
										 nfs_cache->min_free,nfs_sbi->dc_max_size);
		}
		/* FIXGJB: could output more info, and follow /proc/mounts format more closely */
		len += sprintf(buf+len,"\n#  acregmin,max=%d,%d;acdirmin,max=%d,%d;rsize=%d,wsize=%d,bsize=%d\n",
									 nfss->acregmin,nfss->acregmax,nfss->acdirmin,nfss->acdirmax,
									 nfss->rsize, nfss->wsize, nfss->bsize);
		/* do not overflow -- this assume we write no more than 512 bytes per mount --
			 increase the number if you might write more */
		if (len > PAGE_SIZE - 512)
			return len;
	}
	return len;
}

/* a sigusr1 to the nfs fillind process executes this */
void
nfs_profile_empty_data( void )
{
	int i = 0;
	for ( ; i < i_nfs_profile_last_entry; i++) {
		c_nfs_ops[i] = 0;
	}
	cPredHits = 
		cPredUnused =
		cPredCorrect =
		cPredMadeTotal =
		cPredOverflows =
		cRecentUseHit =
		cTotLookups = 0;
}
